package nl.kallestruik.gitlabrepostats

import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.request.*
import io.ktor.http.*
import kotlinx.coroutines.delay
import nl.kallestruik.gitlabrepostats.objects.GitlabIssue
import java.io.File
import kotlin.time.Duration.Companion.seconds

object IssueStats {

    suspend fun printStats(httpClient: HttpClient, gitlabUrl: String, accessToken: String, projectId: String) {
        println("${CC.CYAN}Starting retrieval of issue statistics...${CC.RESET}")

        val apiBaseUrl =
            if (gitlabUrl.endsWith("/"))
                "${gitlabUrl}api/v4/projects/$projectId"
            else
                "$gitlabUrl/api/v4/projects/$projectId"

        var totalPages = -1
        var currentPage = 1

        val issues = mutableListOf<GitlabIssue>()

        do {
            val response = httpClient.get("$apiBaseUrl/issues?per_page=100&page=$currentPage") {
                header("Authorization", "Bearer $accessToken")
            }
            // If we got something other than a 200 OK we are most likely rate-limited.
            if (response.status != HttpStatusCode.OK) {
                delay(1000)
                continue
            }

            // Append the response to the list of jobs.
            issues += response.body<List<GitlabIssue>>()

            totalPages = response.headers["x-total-pages"]?.toIntOrNull() ?: totalPages
            currentPage++
        } while (currentPage <= totalPages)

        val numberOfTotalIssues = issues.size
        val numberOfClosedIssues = issues
            .filter { it.state == "closed" }
            .size

        println("Total number of issues (closed): ${CC.CYAN}$numberOfTotalIssues${CC.RESET} (${CC.RED}$numberOfClosedIssues${CC.RESET})")

        val users = issues
            .flatMap { issue ->
                issue.assignees
                .map { it.name }
            }
            .distinct()

        val issuesByUser = mutableMapOf<String, List<GitlabIssue>>()

        users.forEach { user ->
            issuesByUser[user] = issues
                .filter { issue ->
                    issue.assignees
                        .map { it.name }
                        .contains(user)
                }
        }

        println("Issues per user:")
        issuesByUser.forEach {
                val issueCount = it.value.size
                val closedIssueCount = it.value
                    .filter { it.state == "closed" }
                    .size

                val totalTimeSpent = it.value
                    .sumOf { (it.timeStats.totalTimeSpent ?: 0)  / it.assignees.size }

                val totalTimeEstimated = it.value
                    .sumOf { (it.timeStats.timeEstimate ?: 0)  / it.assignees.size }


                println("${CC.GREEN}${it.key}${CC.RESET}:")
                println("\tTotal (closed): ${CC.CYAN}$issueCount${CC.RESET} (${CC.RED}$closedIssueCount${CC.RESET})")
                println("\tTime spent (estimated):  ${CC.CYAN}${totalTimeSpent.seconds.formatTime()}${CC.RESET} (${CC.GREEN}${totalTimeEstimated.seconds.formatTime()}${CC.RESET})")
            }

        val closedDates = issues
            .mapNotNull { it.closedAt }
            .map { it.toLocalDate() }
            .distinct()
            .sorted()

        val outputFile = File("issue-stats_time-tracking.csv")
        outputFile.createNewFile()

        val userTotals = users.associateWith { 0 }.toMutableMap()

        outputFile.writeText("Date;${users.joinToString(";")}\n")
        closedDates.forEach {date ->
            users.forEach {user ->
                val timeSpentToday = issues
                    .filter { it.assignees.map { it.name }.contains(user) }
                    .filter { it.closedAt?.toLocalDate() == date }
                    .sumOf { (it.timeStats.totalTimeSpent ?: 0) / it.assignees.size }
                userTotals[user] = userTotals[user]!! + timeSpentToday
            }

            outputFile.appendText("$date;${userTotals.map { it.value / 60.0 / 60.0 }.joinToString(";")}\n")
        }
    }
}