package nl.kallestruik.gitlabrepostats

import io.ktor.client.*
import io.ktor.client.call.*
import io.ktor.client.request.*
import io.ktor.http.*
import kotlinx.coroutines.delay
import nl.kallestruik.gitlabrepostats.objects.GitlabJob
import java.text.DecimalFormat
import kotlin.time.Duration
import kotlin.time.Duration.Companion.seconds
import kotlin.time.DurationUnit

object PipelineStats {

    suspend fun printStats(httpClient: HttpClient, gitlabUrl: String, accessToken: String, projectId: String) {
        println("${CC.CYAN}Starting retrieval of pipeline statistics...${CC.RESET}")

        val apiBaseUrl =
            if (gitlabUrl.endsWith("/"))
                "${gitlabUrl}api/v4/projects/$projectId"
            else
                "$gitlabUrl/api/v4/projects/$projectId"

        var totalPages = -1
        var currentPage = 1

        val jobs = mutableListOf<GitlabJob>()

        do {
            val response = httpClient.get("$apiBaseUrl/jobs?per_page=100&page=$currentPage") {
                header("Authorization", "Bearer $accessToken")
            }
            // If we got something other than a 200 OK we are most likely rate-limited.
            if (response.status != HttpStatusCode.OK) {
                delay(1000)
                continue
            }

            // Append the response to the list of jobs.
            jobs += response.body<List<GitlabJob>>()

            totalPages = response.headers["x-total-pages"]?.toIntOrNull() ?: totalPages
            currentPage++
        } while (currentPage <= totalPages)

        val jobsPerBranch = jobs
            .filterNot { it.ref == null }
            .groupBy { it.ref }
            .mapValues { it.value.size }
            .entries.map { Pair(it.key, it.value) }
            .sortedByDescending { it.second }

        val totalJobs = jobs.size

        val failedJobs = jobs.filter { it.status == "failed" }.size
        val failedPercentage = DecimalFormat("#.##").format(failedJobs * 100.0F / totalJobs)

        val totalJobLength = jobs.sumOf { it.duration ?: 0.0 }

        println("${CC.GREEN}## Pipeline Statistics ##${CC.RESET}")
        println("Total job count: ${CC.CYAN}$totalJobs${CC.RESET}")
        println("Of which failed: ${CC.CYAN}$failedJobs ${CC.RESET}($failedPercentage%)")
        println("Time spent executing: ${CC.CYAN}${totalJobLength.seconds.formatTime()}${CC.RESET}")
        println("Branches with most jobs ran:")
        jobsPerBranch.take(10).forEach {
            println("- ${it.first}: ${it.second}")
        }
    }
}